from . import editor
cmd = __import__("sys").modules["pymol.cmd"]
from .cmd import DEFAULT_SUCCESS, DEFAULT_ERROR

# persistent storage between copy/paste
class _PersistentEditing:

    def __init__(self,self_cmd=cmd):
        self._cmd = self_cmd
        self._obj = None

    def get_clipboard_object(self):
        '''name of the temporary persistent object'''
        return self._obj

    def create_tmp(self,sel,extract):
        '''copy or cut selection to clipboard'''
        if self._obj:
            self._cmd.delete(self._obj)
        else:
            self._obj = self._cmd.get_unused_name("_persistent_obj")

        auto_hide_sele = self._cmd.get_setting_boolean("auto_hide_selections")
        self._cmd.set("auto_hide_selections", 0, updates=0)

        try:
            self._cmd.create(self._obj, sel, extract=extract, zoom=0)
            self._cmd.disable(self._obj)
        finally:
            self._cmd.set("auto_hide_selections", auto_hide_sele, updates=0)

# user commands via keyboard

_kCopy  = 0
_kPaste = 1
_kCut   = 2

def editing_ring(action, *, _self=cmd):
    """
DESCRIPTION

    Helper function for copy/cut/paste of molecular selections.

ARGUMENTS

    action = cut/copy/paste/invert
    """

    space = getattr(_self, "_editing_ring_space", None)
    if space is None:
        space = _PersistentEditing(_self)
        _self._editing_ring_space = space

    # PASTE current hidden object into a new object
    if action in (_kPaste, "paste"):
        clipobj = space.get_clipboard_object()
        if not clipobj:
            print("Nothing on clipboard")
        else:
            _self.copy(_self.get_unused_name("obj"), clipobj, zoom=0)
        return

    sels = _self.get_names("public_selections", enabled_only=1)
    if not sels:
        print("No active selection")
        return

    sel = sels[0]
    assert sel

    # COPY current selection into a new hidden object
    if action in (_kCopy, "copy"):
        space.create_tmp(sel, extract=0)

    # CUT current selection into a new hidden object
    elif action in (_kCut, "cut"):
        space.create_tmp(sel, extract=1)

    # INVERT the current selection
    elif action == "invert":
        _self.select(sel, 'not ' + sel)

    else:
        raise AttributeError(action)


def get_default_keys(_self=cmd):
    from pymol.shortcut_manager import shortcut_dict_ref, ShortcutIndex
    keys = {}
    for key in shortcut_dict_ref:
        keys[key] = shortcut_dict_ref[key][ShortcutIndex.COMMAND]

    return keys



#A* -------------------------------------------------------------------
#B* This file contains source code for the PyMOL computer program
#C* Copyright (c) Schrodinger, LLC.
#D* -------------------------------------------------------------------
#E* It is unlawful to modify or remove this copyright notice.
#F* -------------------------------------------------------------------
#G* Please see the accompanying LICENSE file for further information.
#H* -------------------------------------------------------------------
#I* Additional authors of this source file include:
#-*
#-*
#-*
#Z* -------------------------------------------------------------------

if True:

    import sys
    if True:
        import _thread as thread
        import urllib.request as urllib2
        from io import FileIO as file

    import inspect
    import glob
    import shlex
    from enum import Enum
    from functools import wraps
    from pathlib import Path
    from textwrap import dedent
    from typing import List

    import re
    import os
    import time
    import threading
    import traceback
    from . import colorprinting
    from . import parsing
    cmd = sys.modules["pymol.cmd"]
    import pymol

    from .cmd import _cmd, Shortcut, QuietException, \
          fb_module, fb_mask, is_list, \
          DEFAULT_ERROR, DEFAULT_SUCCESS, is_ok, is_error, is_string

    def resume(filename, _self=cmd):
        '''
        
DESCRIPTION

    "resume" executes a log file and opens it for recording of
    additional commands.

USAGE

    resume filename

SEE ALSO

    log, log_close

    '''
        r = DEFAULT_ERROR
        if os.path.exists(filename):
            if(re.search(r"\.py$|\.PY$|\.pym$|.PYM$",filename)):
                r = _self.do("run %s"%filename)
            else:
                r = _self.do("@%s"%filename)
        if is_ok(r):
            r = _self.do("log_open %s,a"%filename)
        if _self._raising(r,_self): raise pymol.CmdException
        return r


    class QueueFile:

        def __init__(self,queue):
            self.queue = queue

        def write(self,command):
            self.queue.put(command)

        def flush(self):
            pass

        def close(self):
            del self.queue

    re_fetch = re.compile(r'(\bfetch\s+\w[^;\r\n\'"]+)')

    class LogFile(file):
        def _append_async0(self, m):
            s = m.group()
            if 'async' in s:
                return s
            return s.rstrip(',') + ', async=0'
        def write(self, s):
            s = re_fetch.sub(self._append_async0, s)
            file.write(self, s.encode())

    def log_open(filename='log.pml', mode='w', _self=cmd):
        '''
DESCRIPTION

    "log_open" opens a log file for writing.

USAGE

    log_open [ filename [, mode ]]

ARGUMENTS

    filename = str: file to write to (.pml or .py) {default: log.pml}

    mode = w/a: "w" to open an empty log file, "a" to append {default: w}

SEE ALSO

    log, log_close
    
        '''
        pymol=_self._pymol
        if not is_string(filename): # we're logging to Queue, not a file.
            pymol._log_file = QueueFile(filename)
            _self.set("logging",1,quiet=1)
        else:
            try:
                try:
                    if hasattr(pymol,"_log_file"):
                        if pymol._log_file is not None:
                            pymol._log_file.close()
                            del pymol._log_file
                except:
                    pass
                pymol._log_file = LogFile(filename,mode)
                if _self._feedback(fb_module.cmd,fb_mask.details): # redundant
                    if mode!='a':
                        print(" Cmd: logging to '%s'."%filename)
                    else:
                        print(" Cmd: appending to '%s'."%filename)
                if mode=='a':
                    pymol._log_file.write("\n") # always start on a new line
                if(re.search(r"\.py$|\.PY$|\.pym$|\.PYM$",filename)):
                    _self.set("logging",2,quiet=1)
                else:
                    _self.set("logging",1,quiet=1)
            except:
                print("Error: unable to open log file '%s'"%filename)
                pymol._log_file = None
                _self.set("logging",0,quiet=1)
                traceback.print_exc()
                raise QuietException

    def log(text, alt_text=None, _self=cmd):
        '''
DESCRIPTION

    "log" writes a command to the log file (if one is open).

    `text` and/or `alt_text` must include the terminating line feed.

ARGUMENTS

    text = str: PyMOL command (optional if alt_text is given)
    alt_text = str: Python expression (optional)

SEE ALSO

    log_open, log_close
    
        '''
        # See also equivalent C impelemtation: PLog

        log_file = getattr(_self._pymol, "_log_file", None)

        if log_file is None:
            return

        mode = _self.get_setting_int("logging")

        if mode == 1:
            # .pml
            if not text and alt_text:
                text = '/' + alt_text
        elif mode == 2:
            # .py
            if alt_text:
                text = alt_text
            elif text.startswith('/'):
                text = text[1:]
            else:
                text = "cmd.do(" + repr(text.strip()) + ")\n"
        else:
            return

        if text:
            log_file.write(text)
            log_file.flush()

    def log_close(_self=cmd):
        '''
DESCRIPTION

    "log_close" closes the current log file (if one is open).

USAGE

    log_close

SEE ALSO

    log, log_open
    
        '''
        pymol=_self._pymol
        cmd=_self
        if hasattr(pymol,"_log_file"):
            if pymol._log_file is not None:
                pymol._log_file.close()
                del pymol._log_file
                _self.set("logging",0,quiet=1)
                if _self._feedback(fb_module.cmd,fb_mask.details): # redundant
                    print(" Cmd: log closed.")

    def cls(_self=cmd):
        '''
DESCRIPTION

    "cls" clears the output buffer.

USAGE

    cls
    '''
        r = DEFAULT_ERROR
        try:
            _self.lock(_self)
            r = _cmd.cls(_self._COb)
        finally:
            _self.unlock(r,_self)
        if _self._raising(r,_self): raise pymol.CmdException
        return r

    def _load_splash_image(filename, url, _self=cmd):
        import tempfile
        import struct

        tmp_filename = ""
        contents = None

        if url:
            try:
                handle = urllib2.urlopen(url)
                contents = handle.read()
                handle.close()

                # png magic number
                if contents[:4] != b'\x89\x50\x4e\x47':
                    raise IOError

                shape = struct.unpack('>II', contents[16:24])

                tmp_filename = tempfile.mktemp('.png')
                with open(tmp_filename, 'wb') as handle:
                    handle.write(contents)

                filename = tmp_filename
            except IOError:
                pass

        if os.path.exists(filename) and not _self.get_names():
            # hide text splash
            print()

            # fit window to image
            try:
                if not contents:
                    contents = open(filename, 'rb').read(24)
                shape = struct.unpack('>II', contents[16:24])
                scale = _self.get_setting_int('display_scale_factor')
                _self.viewport(shape[0] * scale, shape[1] * scale)
            except Exception as e:
                print(e)

            # load image
            _self.load_png(filename, 0, quiet=1)

        if tmp_filename:
            os.unlink(tmp_filename)

    def splash(mode=0, _self=cmd):
        cmd=_self
        '''
DESCRIPTION

    "splash" shows the splash screen information.

USAGE

    splash
        '''
        r = DEFAULT_ERROR
        mode = int(mode)
        if mode == 2: # query
            try:
                _self.lock(_self)
                r = _cmd.splash(_self._COb,1)
            finally:
                _self.unlock(0,_self)
        elif mode == 1: # just show PNG
            show_splash = 1
            try:
                _self.lock(_self)
                show_splash = _cmd.splash(_self._COb,1)
            finally:
                _self.unlock(0,_self)
            r = DEFAULT_SUCCESS
            png_url = ""
            if show_splash==1: # generic / open-source
                png_path = _self.exp_path("$PYMOL_DATA/pymol/splash.png")
            elif show_splash==2: # evaluation builds
                png_path = _self.exp_path("$PYMOL_DATA/pymol/epymol.png")
            elif show_splash==3: # edu builds
                png_path = _self.exp_path("$PYMOL_DATA/pymol/splash_edu.png")
                png_url = "http://pymol.org/splash/splash_edu_2.png"
            else: # incentive builds
                png_path = _self.exp_path("$PYMOL_DATA/pymol/ipymol.png")

            t = threading.Thread(target=_load_splash_image, args=(png_path, png_url, _self))
            t.setDaemon(1)
            t.start()
        else:
            if _self.get_setting_int("internal_feedback") > 0:
                _self.set("text","1",quiet=1)
            print()
            try:
                _self.lock(_self)
                r = _cmd.splash(_self._COb,0)
            finally:
                _self.unlock(r,_self)
        if _self._raising(r,_self): raise pymol.CmdException
        return r

    reinit_code = {
        'everything' : 0,
        'settings' : 1,
        'store_defaults' : 2,
        'original_settings' : 3,
        'purge_defaults' : 4,
    }

    reinit_sc = Shortcut(reinit_code.keys())

    def reinitialize(what='everything', object='', _self=cmd):
        '''
DESCRIPTION

    "reinitialize" reinitializes the program by deleting all objects
    and restoring the default program settings.

USAGE

    reinitialize

        '''
        r = DEFAULT_ERROR
        what = reinit_code[reinit_sc.auto_err(str(what),'option')]
        try:
            _self.lock(_self)
            r = _cmd.reinitialize(_self._COb,int(what),str(object))
        finally:
            _self.unlock(r,_self)
        if _self._raising(r,_self): raise pymol.CmdException
        return r

    def sync(timeout=1.0,poll=0.05,_self=cmd):
        '''
DESCRIPTION

    "sync" is an API-only function which waits until all current
    commmands have been executed before returning.  A timeout
    can be used to insure that this command eventually returns.

PYMOL API

    cmd.sync(float timeout=1.0,float poll=0.05)

SEE ALSO

    frame
    '''
        for t in async_threads:
            t.join()

        now = time.time()
        timeout = float(timeout)
        poll = float(poll)
        # first, make sure there aren't any commands waiting...
        if _cmd.wait_queue(_self._COb): # commands waiting to be executed?
            while 1:
                if not _cmd.wait_queue(_self._COb):
                    break
                e = threading.Event() # using this for portable delay
                e.wait(poll)
                del e
                if (timeout>=0.0) and ((time.time()-now)>timeout):
                    colorprinting.warning('cmd.sync() timed out (wait_queue)')
                    break
        if _cmd.wait_deferred(_self._COb):
            # deferred tasks waiting for a display event?
            if _self.is_gui_thread():
                _self.refresh()
            else:
                while 1:
                    if not _cmd.wait_queue(_self._COb):
                        break
                    e = threading.Event() # using this for portable delay
                    e.wait(poll)
                    del e
                    if (timeout>=0.0) and ((time.time()-now)>timeout):
                        colorprinting.warning('cmd.sync() timed out (wait_deferred)')
                        break
        # then make sure we can grab the API
        while 1:
            if _self.lock_attempt(_self):
                _self.unlock(_self)
                break
            e = threading.Event() # using this for portable delay
            e.wait(poll)
            del e
            if (timeout>=0.0) and ((time.time()-now)>timeout):
                colorprinting.warning('cmd.sync() timed out (lock_attempt)')
                break

    def do(commands,log=1,echo=1,flush=0,_self=cmd):
        # WARNING: don't call this routine if you already have the API lock
        # use cmd._do instead
        '''
DESCRIPTION

    "do" makes it possible for python programs to issue simple PyMOL
    commands as if they were entered on the command line.

PYMOL API

    cmd.do( commands )

USAGE (PYTHON)

    from pymol import cmd
    cmd.do("load file.pdb")
        '''
        log = int(log)
        cmmd_list = commands if is_list(commands) else [commands]
        cmmd_list = [a for cmmd in cmmd_list for a in cmmd.splitlines() if a]
        n_cmmd = len(cmmd_list)
        if n_cmmd>1: # if processing a list of commands, defer updates
            defer = _self.get_setting_int("defer_updates")
            _self.set('defer_updates',1)
        if flush or (_self.is_gui_thread() and _self.lock_api_allow_flush):
            for a in cmmd_list:
                with _self.lockcm:
                    _cmd.do(_self._COb,a,log,echo)
        else:
            with _self.lockcm:
                for a in cmmd_list:
                    _cmd.do(_self._COb,a,log,echo)
        if n_cmmd>1:
            _self.set('defer_updates',defer)

    def quit(code=0, _self=cmd):
        '''
DESCRIPTION

    "quit" terminates the program. 

USAGE

    quit [code]

ARGUMENTS

    code = int: exit the application with status "code" {default: 0}

PYMOL API

    cmd.quit(int code)
        '''
        code = int(code)
        if _self.is_gui_thread():
            _self._quit(code, _self)
        else:
            try:
                _self.lock(_self)
                _cmd.do(_self._COb,"_ time.sleep(0.100);cmd._quit(%d)" % (code),0,0)
                # allow time for a graceful exit from the calling thread
                try:
                    thread.exit()
                except SystemExit:
                    pass
            finally:
                _self.unlock(-1,_self=_self)
        return None

    def delete(name, *, _self=cmd):
        '''
DESCRIPTION

    "delete" removes objects and named selections

USAGE

    delete name

ARGUMENTS

    name = name(s) of object(s) or selection(s), supports wildcards (*)

EXAMPLES

    delete measure*     # delete all objects which names start with "measure"
    delete all          # delete all objects and selections

PYMOL API

    cmd.delete (string name = object-or-selection-name )

SEE ALSO

    remove
    delete_states
        '''
        r = DEFAULT_ERROR
        try:
            _self.lock(_self)
            r = _cmd.delete(_self._COb,str(name))
        finally:
            _self.unlock(r,_self)
        if _self._raising(r,_self): raise pymol.CmdException
        return r

    def delete_states(name: str, states: str, *, _self=cmd):
        '''
DESCRIPTION

USAGE

    delete_states name, states

ARGUMENTS

    name = name(s) of object(s), supports wildcards (*)

    states = string: space separated list of state numbers or ranges

EXAMPLES

    delete_state 1nmr, 1-5     # delete states 1 to 5 from 1nmr
    delete_state *, 1-3 10-40  # deletes states 1 to 3 and 10 to 40 from all applicable objects

PYMOL API

    cmd.delete_states(string name = object-name, string states = states string)

SEE ALSO

    delete
        '''

        def get_state_list(states_str):
            output=[]
            input_str = re.sub(r"\s"," ", states_str)
            input_str = input_str.replace("-"," -")
            input_str = input_str.replace("- ","-")
            input_str = input_str.strip().split()

            last = -1
            for x in input_str:
                if not x.isdigit():
                    if x.startswith("-"):
                        direction = 1
                        current = last
                        last = int(x[1:]) - 1
                        if last < current:
                            direction = -1
                        while current != last:
                            current += direction
                            output.append(str(current))
                else:
                    val = int(x) - 1
                    output.append(str(val))
                    last = val
            return output

        with _self.lockcm:
            output = get_state_list(states)
            states_list = sorted(set(map(int, output)))
            return _cmd.delete_states(_self._COb, name, states_list)

    class Selection(str):
        pass


    def _parse_bool(value: str):
        if isinstance(value, str):
            if value.lower() in ["yes", "1", "true", "on", "y"]:
                return True
            elif value.lower() in ["no", "0", "false", "off", "n"]:
                return False
            else:
                raise Exception("Invalid boolean value: %s" % value)
        elif isinstance(value, bool):
            return value
        else:
            raise Exception(f"Unsuported boolean flag {value}")

    def _parse_list_str(value):
        return shlex.split(value)

    def _parse_list_int(value):
        return list(map(int, shlex.split(value)))

    def _parse_list_float(value):
        return list(map(float, shlex.split(value)))

    def declare_command(name, function=None, _self=cmd):
        if function is None:
            name, function = name.__name__, name

        # new style commands should have annotations
        annotations = [a for a in function.__annotations__ if a != "return"]
        if function.__code__.co_argcount != len(annotations):
            raise Exception("Messy annotations")

        # docstring text, if present, should be dedented
        if function.__doc__ is not None:
            function.__doc__ = dedent(function.__doc__).strip()


        # Analysing arguments
        spec = inspect.getfullargspec(function)
        kwargs_ = {}
        args_ = spec.args[:]
        defaults = list(spec.defaults or [])

        args2_ = args_[:]
        while args_ and defaults:
            kwargs_[args_.pop(-1)] = defaults.pop(-1)

        funcs = {}
        for idx, (var, func) in enumerate(spec.annotations.items()):
            funcs[var] = func

        # Inner function that will be callable every time the command is executed
        @wraps(function)
        def inner(*args, **kwargs):
            frame = traceback.format_stack()[-2]
            caller = frame.split("\"", maxsplit=2)[1]

            # It was called from command line or pml script, so parse arguments
            if caller.endswith("pymol/parser.py"):
                kwargs = {**kwargs_, **kwargs, **dict(zip(args2_, args))}
                kwargs.pop("_self", None)
                for arg in kwargs.copy():
                    if funcs[arg] == bool:
                        funcs[arg] = _parse_bool
                    elif funcs[arg] == List[str]:
                        funcs[arg] = _parse_list_str
                    elif funcs[arg] == List[int]:
                        funcs[arg] = _parse_list_int
                    elif funcs[arg] == List[float]:
                        funcs[arg] = _parse_list_float
                    else:
                        # Assume it's a literal supported type
                        pass
                    # Convert the argument to the correct type
                    kwargs[arg] = funcs[arg](kwargs[arg])
                return function(**kwargs)

            # It was called from Python, so pass the arguments as is
            else:
                return function(*args, **kwargs)

        name = function.__name__
        _self.keyword[name] = [inner, 0, 0, ",", parsing.STRICT]
        _self.kwhash.append(name)
        _self.help_sc.append(name)
        return inner

    def extend(name, function=None, _self=cmd):

        '''
DESCRIPTION

    "extend" is an API-only function which binds a new external
    function as a command into the PyMOL scripting language.

PYMOL API

    cmd.extend(string name,function function)

PYTHON EXAMPLE

    def foo(moo=2): print moo
    cmd.extend('foo',foo)

    The following would now work within PyMOL:

    PyMOL>foo
    2
    PyMOL>foo 3
    3
    PyMOL>foo moo=5
    5
    PyMOL>foo ?
    Usage: foo [ moo ]

NOTES

    For security reasons, new PyMOL commands created using "extend" are
    not saved or restored in sessions.

SEE ALSO

    alias, api
            '''
        if function is None:
            name, function = name.__name__, name
        _self.keyword[name] = [function, 0,0,',',parsing.STRICT]
        _self.kwhash.append(name)
        _self.help_sc.append(name)
        return function

        # for aliasing compound commands to a single keyword

    def extendaa(*arg, _self=cmd):
        '''
DESCRIPTION

    API-only function to decorate a function as a PyMOL command with
    argument auto-completion.

EXAMPLE

    @cmd.extendaa(cmd.auto_arg[0]['zoom'])
    def zoom_organic(selection='*'):
        cmd.zoom('organic & (%s)' % selection)
        '''
        auto_arg = _self.auto_arg
        def wrapper(func):
            name = func.__name__
            _self.extend(name, func)
            for (i, aa) in enumerate(arg):
                if i == len(auto_arg):
                    auto_arg.append({})
                if aa is not None:
                    auto_arg[i][name] = aa
            return func
        return wrapper

    def alias(name, command, _self=cmd):
        '''
DESCRIPTION

    "alias" binds routinely-used command inputs to a new command
    keyword.

USAGE

    alias name, command

ARGUMENTS

    name = string: new keyword
    
    command = string: literal input with commands separated by semicolons.
    
EXAMPLE

    alias my_scene, hide; show ribbon, polymer; show sticks, organic; show nonbonded, solvent
    
    my_scene

NOTES

    For security reasons, aliased commands are not saved or restored
    in sessions.

SEE ALSO

    cmd.extend, api
            '''
        _self.keyword[name] = [eval("lambda :do('''%s ''')"%command.replace("'''","")),
                               0,0,',',parsing.STRICT]
        _self.kwhash.append(name)

    async_threads = []

    def async_(func, *args, _self=cmd, **kwargs):
        '''
DESCRIPTION

    Run function threaded and show "please wait..." message.
        '''
        from .wizard.message import Message

        wiz = Message(['please wait ...'], dismiss=0, _self=_self)

        try:
            _self.set_wizard(wiz)
        except:
            wiz = None

        if isinstance(func, str):
            func = _self.keyword[func][0]

        def wrapper():
            async_threads.append(t)
            try:
                func(*args, **kwargs)
            except (pymol.CmdException, cmd.QuietException) as e:
                if e.args:
                    print(e)
            finally:
                if wiz is not None:
                    try:
                        _self.set_wizard_stack([w
                            for w in _self.get_wizard_stack() if w != wiz])
                    except:
                        _self.do('_ wizard')
                    else:
                        _self.refresh_wizard()

                async_threads.remove(t)

        t = threading.Thread(target=wrapper)
        t.setDaemon(1)
        t.start()



"""
Wrapper for PyMOL scripts to get PySide or PyQt

Useful link for PySide/PyQt4 differences:
https://deptinfo-ensip.univ-poitiers.fr/ENS/pyside-docs/pysideapi2.html

PyQt5/PyQt4 differences:
http://pyqt.sourceforge.net/Docs/PyQt5/pyqt4_differences.html
"""

DEBUG = False

PYQT_NAME = None
QtWidgets = None

try:
    from pymol._Qt_pre import *
except ImportError:
    if DEBUG:
        print('import _Qt_pre failed')

PYQT_NAME = None
QtWidgets = None

try:
    from pymol._Qt_pre import *
except ImportError:
    if DEBUG:
        print('import _Qt_pre failed')

import os

qt_api = os.environ.get('QT_API', '')

if not PYQT_NAME and qt_api in ('', 'pyqt5'):
    try:
        from PyQt5 import QtGui, QtCore, QtOpenGL, QtWidgets
        PYQT_NAME = 'PyQt5'
    except ImportError:
        if DEBUG:
            print('import PyQt5 failed')

if not PYQT_NAME and qt_api in ('', 'pyside2'):
    try:
        from PySide2 import QtGui, QtCore, QtOpenGL, QtWidgets
        PYQT_NAME = 'PySide2'
    except ImportError:
        if DEBUG:
            print('import PySide2 failed')

if not PYQT_NAME and qt_api in ('', 'pyqt6'):
    try:
        from PyQt6 import QtGui, QtCore, QtOpenGL, QtWidgets
        from PyQt6 import QtOpenGLWidgets
        PYQT_NAME = 'PyQt6'
    except ImportError:
        if DEBUG:
            print('import PyQt6 failed')

if not PYQT_NAME and qt_api in ('', 'pyside6'):
    try:
        from PySide6 import QtGui, QtCore, QtOpenGL, QtWidgets
        from PySide6 import QtOpenGLWidgets
        PYQT_NAME = 'PySide6'
    except ImportError:
        if DEBUG:
            print('import PySide6 failed')

if not PYQT_NAME:
    raise ImportError(__name__)

# qtpy compatibility
os.environ['QT_API'] = PYQT_NAME.lower()

if QtWidgets is None:
    QtWidgets = QtGui

if hasattr(QtCore, 'QAbstractProxyModel'):
    QtCoreModels = QtCore
else:
    QtCoreModels = QtGui

if PYQT_NAME.endswith('6'):
    QtWidgets.QOpenGLWidget = QtOpenGLWidgets.QOpenGLWidget
    QtWidgets.QActionGroup = QtGui.QActionGroup
    QtWidgets.QAction = QtGui.QAction
    QtWidgets.QShortcut = QtGui.QShortcut
    QtCore.QSortFilterProxyModel.setFilterRegExp = QtCore.QSortFilterProxyModel.setFilterRegularExpression
    QtGui.QFont.Monospace = QtGui.QFont.StyleHint.Monospace

    def copy_attributes(target_class, source_class):
        for attr in dir(source_class):
            if not attr.startswith('_'):
                setattr(target_class, attr, getattr(source_class, attr))

    copy_attributes(QtCore.QEvent, QtCore.QEvent.Type)
    copy_attributes(QtCore.Qt, QtCore.Qt.AlignmentFlag)
    copy_attributes(QtCore.Qt, QtCore.Qt.CaseSensitivity)
    copy_attributes(QtCore.Qt, QtCore.Qt.CheckState)
    copy_attributes(QtCore.Qt, QtCore.Qt.ContextMenuPolicy)
    copy_attributes(QtCore.Qt, QtCore.Qt.DockWidgetArea)
    copy_attributes(QtCore.Qt, QtCore.Qt.FocusPolicy)
    copy_attributes(QtCore.Qt, QtCore.Qt.GestureType)
    copy_attributes(QtCore.Qt, QtCore.Qt.ItemFlag)
    copy_attributes(QtCore.Qt, QtCore.Qt.Key)
    copy_attributes(QtCore.Qt, QtCore.Qt.KeyboardModifier)
    copy_attributes(QtCore.Qt, QtCore.Qt.MouseButton)
    copy_attributes(QtCore.Qt, QtCore.Qt.Orientation)
    copy_attributes(QtCore.Qt, QtCore.Qt.WindowType)
    copy_attributes(QtGui.QFont, QtGui.QFont.StyleHint)
    copy_attributes(QtWidgets.QAbstractItemView, QtWidgets.QAbstractItemView.ScrollHint)
    copy_attributes(QtWidgets.QAbstractItemView, QtWidgets.QAbstractItemView.SelectionBehavior)
    copy_attributes(QtWidgets.QAbstractItemView, QtWidgets.QAbstractItemView.SelectionMode)
    copy_attributes(QtWidgets.QBoxLayout, QtWidgets.QBoxLayout.Direction)
    copy_attributes(QtWidgets.QMainWindow, QtWidgets.QMainWindow.DockOption)
    copy_attributes(QtWidgets.QOpenGLWidget, QtOpenGLWidgets.QOpenGLWidget.UpdateBehavior)
    copy_attributes(QtWidgets.QSizePolicy, QtWidgets.QSizePolicy.Policy)
    copy_attributes(QtWidgets.QTreeWidgetItem, QtWidgets.QTreeWidgetItem.ChildIndicatorPolicy)

    QtCore.Qt.MidButton = QtCore.Qt.MiddleButton
    QtCore.Qt.WA_LayoutUsesWidgetRect = QtCore.Qt.WidgetAttribute.WA_LayoutUsesWidgetRect


if PYQT_NAME[:4] == 'PyQt':
    QtCore.Signal = QtCore.pyqtSignal
    QtCore.Slot = QtCore.pyqtSlot
else:
    QtCore.pyqtSignal = QtCore.Signal
    QtCore.pyqtSlot = QtCore.Slot
    QtCore.QT_VERSION_STR = QtCore.__version__
    QtCore.QT_VERSION = (
            0x10000 * QtCore.__version_info__[0] +
            0x00100 * QtCore.__version_info__[1] +
            0x00001 * QtCore.__version_info__[2])

del qt_api
del os



import os

from pymol.Qt import QtGui, QtCore, QtWidgets
from pymol.Qt.utils import UpdateLock, PopupOnException
import pymol
Qt = QtCore.Qt

from pymol.setting import name_dict

# Functions to convert atom property values to strings for display in the GUI
strfunctions = {
    'color': lambda c: hex(c) if (c >= 0x40000000) else str(c),
    'reps': bin,
    'flags': bin,
}

class UneditableDelegate(QtWidgets.QStyledItemDelegate):
    def createEditor(self, parent, option, index):
        return None

class FunctionSuspender:
    def __init__(self, func):
        self.func = func
    def __enter__(self):
        self.func.suspended = True
    def __exit__(self, exc_type, exc_value, traceback):
        self.func.suspended = False

def suspendable(func):
    def wrapper(*args, **kwargs):
        if not func.suspended:
            return func(*args, **kwargs)
    func.suspended = False
    wrapper.suspend = FunctionSuspender(func)
    return wrapper

def get_object_names(_self):
    # was _self.get_names('public_objects') in PyMOL 2.1
    # but that throws exceptions for groups/isosurfaces/etc.
    names = _self.get_object_list()
    return names

class PropsDialog(QtWidgets.QWidget):
    def __init__(self, parent):
        QtWidgets.QWidget.__init__(self)
        self.cmd = parent.cmd
        self.form = parent.load_form('props')
        self.setup_tree_widget()
        self.setup_behavior()
        self.item_changed_skip = False

    def make_entry(self, parent, label):
        item = QtWidgets.QTreeWidgetItem(parent)
        item.setText(0, str(label))
        item.setFlags(QtCore.Qt.ItemIsEditable |
                      QtCore.Qt.ItemIsEnabled |
                      QtCore.Qt.ItemIsSelectable )
        return item

    def make_cat(self, parent, label):
        item = QtWidgets.QTreeWidgetItem(parent)
        item.setText(0, str(label))
        item.setFirstColumnSpanned(True)
        item.setExpanded(True)
        item.setChildIndicatorPolicy(
            QtWidgets.QTreeWidgetItem.ShowIndicator)
        return item

    def setup_tree_widget(self):
        self.form.treeWidget.setItemDelegateForColumn(0, UneditableDelegate())

        self.item_object = self.make_cat(self.form.treeWidget, "Object-Level")
        self.item_object_ttt = self.make_entry(self.item_object, "TTT Matrix")
        self.item_object_settings = self.make_cat(self.item_object, "Settings")

        self.item_ostate = self.make_cat(self.form.treeWidget, "Object-State-Level")
        self.item_ostate_title = self.make_entry(self.item_ostate, "Title")
        self.item_ostate_matrix = self.make_entry(self.item_ostate, "State Matrix")
        self.item_ostate_settings = self.make_cat(self.item_ostate, "Settings")
        self.item_ostate_properties = Ellipsis  # Incentive PyMOL only

        self.item_atom = self.make_cat(self.form.treeWidget, "Atom-Level")
        self.item_atom_identifiers = self.make_cat(self.item_atom, "Identifiers")
        self.item_atom_builtins = self.make_cat(self.item_atom, "Properties (built-in)")
        self.item_atom_settings = self.make_cat(self.item_atom, "Settings")
        self.item_atom_properties = Ellipsis  # Incentive PyMOL only

        self.item_astate = self.make_cat(self.form.treeWidget, "Atom-State-Level")
        self.item_astate_builtins = self.make_cat(self.item_astate, "Properties (built-in)")
        self.item_astate_settings = self.make_cat(self.item_astate, "Settings")

        self.keys_atom_identifiers = ['model', 'index', 'segi', 'chain', 'resi',
                                'resn',
                                'oneletter',  # read-only
                                'name', 'alt', 'ID', 'rank']
        self.keys_atom_builtins = ['elem', 'q', 'b', 'type', 'formal_charge',
                            'partial_charge', 'numeric_type', 'text_type',
                            # avoid stereo auto-assignment errors
                            # 'stereo',
                            'vdw', 'ss', 'color', 'reps',
                            'flags',
                            'label', 'cartoon',
                            'protons', 'geom', 'valence', 'elec_radius']
        self.keys_astate_builtins = ['state', 'x', 'y', 'z']

        self.items = {}
        for key in self.keys_atom_identifiers:
            self.items[key] = self.make_entry(self.item_atom_identifiers, key)
        for key in self.keys_atom_builtins:
            self.items[key] = self.make_entry(self.item_atom_builtins, key)
        for key in self.keys_astate_builtins:
            self.items[key] = self.make_entry(self.item_astate_builtins, key)

        self.items['model'].setDisabled(True)
        self.items['index'].setDisabled(True)
        self.items['state'].setDisabled(True)
        self.items['oneletter'].setDisabled(True)

    def setup_behavior(self):
        # init input fields
        self.form.input_model.addItems(get_object_names(self.cmd))
        self.form.input_state.setValue(self.cmd.get_state())

        # select pk1 atom if available
        self.update_from_pk1()

        # Install an eventfilter
        self.form.treeWidget.installEventFilter(self)

        # hook up events
        self.form.input_model.currentIndexChanged.connect(self.update_treewidget_model)
        self.form.input_state.valueChanged.connect(self.update_treewidget_state)
        self.form.input_index.valueChanged.connect(self.update_treewidget)
        self.form.button_refresh.clicked.connect(self.update_model_list)

        # themed icons only available by default on X11
        if self.form.button_refresh.icon().isNull():
            self.form.button_refresh.setIcon(QtGui.QIcon(
                os.path.expandvars('$PYMOL_DATA/pmg_qt/icons/refresh.svg')))

        # update and show
        self.update_treewidget_model()
        self.form.treeWidget.setColumnWidth(0, 200)

        self.form.treeWidget.itemChanged.connect(self.item_changed)

    def get_dialog(self):
        return self.form._dialog

    def item_changed(self, item, column):
        """
        Edits current item.
        """

        if self.item_changed_skip:
            return

        model = self.form.input_model.currentText()
        state = self.form.input_state.value()
        key = item.text(0)
        new_value = item.text(1)
        parent = item.parent()
        result = False

        if not result:
            if item is self.item_object_ttt:
                try:
                    if new_value:
                        result = self.cmd.set_object_ttt(model, new_value)
                except (ValueError, IndexError):
                    result = False
            elif item is self.item_ostate_title:
                result = self.cmd.set_title(model, state, new_value)
            elif item is self.item_ostate_matrix:
                self.cmd.matrix_reset(model, state)
                try:
                    new_value = self.cmd.safe_eval(new_value)
                    result = self.cmd.transform_object(model, new_value, state)
                except: # CmdTransformObject-DEBUG: bad matrix
                    result = False
            elif parent is self.item_object_settings:
                with PopupOnException():
                    self.cmd.set(key, new_value, model, quiet=0)
            elif parent is self.item_ostate_settings:
                with PopupOnException():
                    self.cmd.set(key, new_value, model, state, quiet=0)
            elif parent is self.item_ostate_properties:
                self.cmd.set_property(key, new_value, model, state, quiet=0)
            else:
                is_state = False

                if parent is self.item_atom_properties:
                    key = 'p.' + key
                elif parent is self.item_atom_settings:
                    key = 's.' + key
                elif parent is self.item_astate_settings:
                    key = 's.' + key
                    is_state = True
                elif key in self.keys_astate_builtins:
                    is_state = True

                def get_new_value(old_value):
                    if isinstance(old_value, (tuple, list, bool)):
                        return self.cmd.safe_eval(new_value)
                    try:
                        # cast to old type (required for e.g. 'resv = "3"')
                        if isinstance(old_value, int):
                            return int(new_value, 0)
                        return type(old_value)(new_value)
                    except ValueError:
                        # return str and let PyMOL handle it (e.g. change
                        # type of user property)
                        return new_value.encode('utf-8')

                alter_args = ('pk1', key + '= get_new_value(' + key + ')', 0,
                        {'get_new_value': get_new_value})

                with PopupOnException():
                    if is_state:
                        result = self.cmd.alter_state(state, *alter_args)
                    else:
                        result = self.cmd.alter(*alter_args)

        if not result:
            self.update_treewidget_model()

        self.item_changed_skip = False

    def unset_item(self, item):
        '''
        Calls unset on item if applicable. Matrices will be reset.
        '''
        model = self.form.input_model.currentText()
        state = self.form.input_state.value()
        key = item.text(0)
        old_value = item.text(1)
        old_value_type = type(old_value)
        parent = item.parent()

        unset_result = True

        if old_value == "":
            unset_result = False
            return unset_result

        if item is self.item_object_ttt:
            self.cmd.matrix_reset(model, mode=1) # mode=1 for TTT matrix reset
        elif item is self.item_ostate_title:
            self.cmd.set_title(model, state, '')
        elif item is self.item_ostate_matrix:
            self.cmd.matrix_reset(model, state, mode=2) # mode=2 for state matrix reset
        elif parent is self.item_object_settings:
            self.cmd.unset(key, model, quiet=0)
        elif parent is self.item_ostate_settings:
            self.cmd.unset(key, model, state, quiet=0)
        elif parent is self.item_ostate_properties:
            self.cmd.set_property(key, None, model, state, quiet=0)
        elif parent is self.item_atom_properties:
            key = 'p.' + key
            alter_args = ('pk1', key + '= None', 0)
            unset_result = self.cmd.alter(*alter_args)
        elif parent is self.item_atom_settings:
            key = 's.' + key
            alter_args = ('pk1', key + '= None', 0)
            unset_result = self.cmd.alter(*alter_args)
        else:
            unset_result = False

        if unset_result:
            self.update_treewidget_model()

        return unset_result

    def unset_caller(self):
        item_list_selected = self.form.treeWidget.selectedItems()
        self.unset_item(item_list_selected[0])

    def eventFilter(self, source, event):
        '''
        Event filter for creating new shortcuts. Processes the key event before passing it on.
        '''
        if (event.type() == QtCore.QEvent.KeyPress and source is self.form.treeWidget):
            if (event.key() == QtCore.Qt.Key_Delete):
                self.unset_caller()
                return 0
        return super().eventFilter(source, event)

    def update_object_settings(self, parent, model, state):
        parent.takeChildren()
        for sitem in (self.cmd.get_object_settings(model, state) or []):
            key = name_dict.get(sitem[0], sitem[0])
            item = self.make_entry(parent, key)
            item.setText(1, str(sitem[2]))

    def update_atom_settings(self, wrapper, parent):
        parent.takeChildren()
        for key in wrapper:
            item = self.make_entry(parent, name_dict.get(key, key))
            value = wrapper[key]
            item.setText(1, str(value))

    def update_atom_properties(self, wrapper, parent):
        parent.takeChildren()
        for key in wrapper:
            item = self.make_entry(parent, key)
            value = wrapper[key]
            item.setText(1, str(value))

    def update_atom_fields(self, ns):
        for key in self.keys_atom_identifiers + self.keys_atom_builtins:
            try:
                value = ns[key]
            except Exception as e:
                value = 'ERROR: ' + str(e)
            self.items[key].setText(1, strfunctions.get(key, str)(value))
        self.update_atom_settings(ns['s'], self.item_atom_settings)

    def update_astate_fields(self, ns):
        for key in self.keys_astate_builtins:
            value = ns[key]
            self.items[key].setText(1, str(value))
        self.update_atom_settings(ns['s'], self.item_astate_settings)

    def update_from_pk1(self):
        pk1_atom = []
        if self.cmd.iterate('?pk1', 'pk1_atom[:] = [model, index]', space=locals()) > 0:
            self.form.input_model.setCurrentIndex(self.form.input_model.findText(pk1_atom[0]))
            self.form.input_index.setValue(pk1_atom[1])

    def update_pk1(self):
        model = self.form.input_model.currentText()
        index = self.form.input_index.value()

        if model and index:
            try:
                self.cmd.edit((model, index))
                return True
            except pymol.CmdException:
                pass

        return False

    def update_treewidget(self, *args):
        if not self.update_pk1():
            return

        state = self.form.input_state.value()

        self.item_changed_skip = True
        space = {'update_atom_fields': self.update_atom_fields,'update_astate_fields': self.update_astate_fields,'locals': locals}
        count = self.cmd.iterate(
            'pk1', 'update_atom_fields(locals())', space=space)
        self.item_atom.setDisabled(not count)
        if count:
            count = self.cmd.iterate_state(
                state, 'pk1',
                'update_astate_fields(locals())', space=space)
        self.item_astate.setDisabled(not count)
        self.item_changed_skip = False

    def update_treewidget_state(self, *args):
        model = self.form.input_model.currentText()
        state = self.form.input_state.value()
        if not (model and state):
            return

        self.item_changed_skip = True
        self.item_ostate_title.setText(
            1, str(self.cmd.get_title(model, state) or ''))
        self.item_ostate_matrix.setText(
            1, str(
                self.cmd.get_object_matrix(
                    model, state, 0) or ''))

        self.update_object_settings(self.item_ostate_settings, model, state)

        self.update_treewidget()
        self.item_changed_skip = False

    @suspendable
    def update_treewidget_model(self, *args):
        self.item_changed_skip = True
        model = self.form.input_model.currentText()

        if not model:
            return

        self.item_object_ttt.setText(1, str(self.cmd.get_object_ttt(model) or ''))
        self.update_object_settings(self.item_object_settings, model, 0)

        natoms = self.cmd.count_atoms('?' + model)
        nstates = self.cmd.count_states('?' + model)

        self.form.input_state.setMinimum(1)
        self.form.input_state.setMaximum(nstates)
        self.form.input_index.setMinimum(1)
        self.form.input_index.setMaximum(natoms)

        self.item_atom.setHidden(natoms == 0)
        self.item_astate.setHidden(natoms == 0)
        self.item_ostate.setHidden(nstates == 0)

        self.update_treewidget_state()
        self.item_changed_skip = False

    def update_model_list(self, *args):
        if 'pk1' not in self.cmd.get_names('selections'):
            self.update_pk1()

        with self.update_treewidget_model.suspend:
            self.form.input_model.clear()
            self.form.input_model.addItems(get_object_names(self.cmd))
            self.update_from_pk1()

        self.update_treewidget_model()
